local mod = LibStub("AceAddon-3.0"):NewAddon("BigBrother_Ready", "AceEvent-3.0", "AceTimer-3.0")

--[===[@debug@
_G.BBReady = mod

local PREFIX = "|cff00ff00BB_Ready|r: "
local DevTools_Dump
DevTools_Dump = function(value)
	if not _G.DevTools_Dump then
		LoadAddOn("Blizzard_DebugTools")
	end
	DevTools_Dump = _G.DevTools_Dump
	DevTools_Dump(value)
end

local function dump(value)
	local orig = DEFAULT_CHAT_FRAME
	DEFAULT_CHAT_FRAME = ChatFrame4
	DevTools_Dump(value)
	DEFAULT_CHAT_FRAME = orig
end

local function mystrjoin(arg1, ...)
	local text = tostring(arg1)
	for i = 1, select('#', ...) do
		text = text .. tostring(select(i, ...))
	end
	return text
end

local doDebug
local function realDebug(arg1, ...)
	if not arg1 then return end
	if type(arg1) == 'table' then
		ChatFrame4:AddMessage(PREFIX .. "+++ table-dump")
		dump(arg1)
		ChatFrame4:AddMessage(PREFIX .. "--- end of table-dump")
		realDebug(...)
	else
		local text = mystrjoin(PREFIX, arg1, ...)
		ChatFrame4:AddMessage(text)
	end
end

local buffer = {}
local function collectDebug(...)
	buffer[#buffer + 1] = { ... }
end
doDebug = collectDebug

local function releaseDebug(self)
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	doDebug = realDebug
	for k,v in ipairs(buffer) do
		doDebug(unpack(v))
	end
	wipe(buffer)
	buffer = nil
end
local DebugFrame = CreateFrame('Frame')
DebugFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
DebugFrame:SetScript("OnEvent", releaseDebug)
--@end-debug@]===]

local function debug(...)
	--[===[@debug@
	doDebug(...)
	--@end-debug@]===]
end
mod.debug = debug

local foods = {
	57356, -- Rhinolicious Wormsteak
	57358, -- Hearty Rhino
	57360, -- Snapper Extreme/Worg Tartare
	57367, -- Blackened Dragonfin
	57365, -- Cuttlesteak
	57371, -- Dragonfin Filet
	57399, -- Fish Feast
	57329, -- Spiced Worm Burger/Spicy Blue Nettlefish
	57332, -- Imperial Manta Steak/Very Burnt Worg
	57327, -- Firecracker Salmon/Tender Shoveltusk Steak
	57325, -- Mega Mammoth Meal/Poached Northern Sculpin
	57334, -- Mighty Rhino Dogs/Spicy Fried Herring
}

local flasks = {
	-- Flask WotLK
	53752, -- 53752 Lesser Flask of Toughness
	53755, -- 53755 Flask of the Frost Wyrm
	53758, -- 53758 Flask of Stoneblood
	53760, -- 53760 Flask of Pure Mojo
	54212, -- 54212 Flask of Endless Rage
	62380, -- Lesser Flask of Resistance
--	67016, -- Flask of North
}

local guardianElixirs = {
	53747, -- 53747 Elixir of Spirit
	60347, -- 60347 Elixir of Mighty Thoughts
	53764, -- 53764 Elixir of Mighty Mageblood
	53751, -- 53751 Elixir of Mighty Fortitude
	60343, -- 60343 Elixir of Mighty Defense
}

local battleElixirs = {
	53746, -- 53746 Wrath Elixir
	53749, -- 53749 Guru's Elixir
	53763, -- 53763 Elixir of Protection
	53748, -- 53748 Elixir of Mighty Strength
	28497, -- 53748 Elixir of Mighty Agility
	60346, -- 60346 Elixir of Lightning Speed
	60344, -- 60344 Elixir of Expertise
	60341, -- 60341 Elixir of Deadly Strikes
	60345, -- 60345 Elixir of Armor Piercing
	60340, -- 60340 Elixir of Accuracy
	33721, -- 33721 Spellpower Elixir
}

local db
local timers = {}

local function checkForAnyBuff(unit, buffList)
	for i,v in ipairs(buffList) do
		local spellName = GetSpellInfo(v)
        local name, _, _, _, _, duration, expirationTime,_,_,_,spellid = UnitAura(unit, spellName)
        if name and spellid == v then
			if expirationTime then
				local remaining = expirationTime - GetTime()
				return remaining
			else
				return
			end
		end
	end
	return
end

local function checkForMissingElixir(unit)
	local MIN_DURATION = db.threshold
	local guardianRemaining = checkForAnyBuff(unit, guardianElixirs)
	local battleRemaining = checkForAnyBuff(unit, battleElixirs)
	if guardianRemaining and battleRemaining then
		if guardianRemaining > MIN_DURATION and battleRemaining > MIN_DURATION then
			return
		elseif guardianRemaining < MIN_DURATION or battleRemaining < MIN_DURATION then
			return "Expiring elixirs"
		end
	end
	local onlybuff = guardianRemaining or battleRemaining
	if onlybuff then
		if onlybuff < MIN_DURATION then
			return "Only one expiring elixir"
		else
			return "Only one elixir"
		end
	else
		return "No flask or elixir"
	end
	return false
end

local function checkForMissingFlask(unit)
	local remaining = checkForAnyBuff(unit, flasks)
	if not remaining then
		return checkForMissingElixir(unit)
	elseif remaining < db.threshold then
		return remaining
	else
		return
	end
end

local function checkForMissingFood(unit)
	local remaining = checkForAnyBuff(unit, foods)
	if remaining then
		if remaining < db.threshold then
			return remaining
		else
			return
		end
	elseif UnitAura(unit, GetSpellInfo(foods[1])) then
		return "Low food buff"
	else
		return "No food buff"
	end
end


local function broadCastToRaid(prefix, table)
	if #table <= 0 then
		return
	end
	local message = prefix .. ": " .. strjoin(", ", unpack(table))
	if db.channel == "SELF" then
		print("BBReady: ", message)
	elseif db.channel == "CUSTOM" then
		-- NYI
	else
		SendChatMessage(prefix .. ": " .. strjoin(", ", unpack(table)), db.channel)
	end
end
local TOKEN = "<BB_Ready>"
local outgoingFilter = function(self, event, msg, ...)
	if msg:find(TOKEN) then return true end
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", outgoingFilter)

local function broadCastWhisper(state, table)
	for _, v in ipairs(table) do
		SendChatMessage(("%s %s"):format(TOKEN, state), "WHISPER", nil, v)
	end
end

local function autoInitWithEmptyTable(self, key)
	local val = rawget(self, key)
	if not val then
		val = {}
		rawset(self, key, val)
	end
	return val
end
local mt = { __index = autoInitWithEmptyTable, }

local function getRelevantGroups()
	if db.groups.auto then
		local difficulty = GetRaidDifficulty()
		if difficulty%2 == 1 then
			return { true, true, }
		else
			return { true, true, true, true, true, }
		end
	else
		return CopyTable(db.groups)
	end
end

local function formatTime(seconds)
	local minutes = math.floor(seconds/60)
	local seconds = math.floor(seconds - minutes * 60)
	if seconds < 10 then
		seconds = "0" .. seconds
	end
	return ("%s:%s"):format(minutes, seconds)
end

function mod:READY_CHECK()
	local buffStates = setmetatable({}, mt)
	local groups = getRelevantGroups()
	for i = 1, GetNumRaidMembers() do
		local unit = "raid" .. i
		local name = UnitName(unit)
		local _, _, group = GetRaidRosterInfo(i)
		if not groups[group] then
			-- nothing
		elseif not UnitIsVisible(unit) then
			table.insert(buffStates["out of range"], name)
		else
			local state = checkForMissingFlask(unit)
			if type(state) == 'number' then
				table.insert(buffStates["Expiring Flask"], ("%s (%s)"):format(name, formatTime(state)))
			elseif state then
				table.insert(buffStates[state], name)
			end
			local foodstate = checkForMissingFood(unit)
			if type(foodstate) == 'number' then
				table.insert(buffStates["Expiring foodbuff"], ("%s (%s)"):format(name, formatTime(foodstate)))
			elseif foodstate then
				table.insert(buffStates[foodstate], name)
			end
		end
	end
	for k,v in pairs(buffStates) do
		broadCastToRaid(k, v)
	end
	if db.state == 2 then
		wipe(buffStates["out of range"])
		for k,v in pairs(buffStates) do
			broadCastWhisper(k, v)
		end
	end
	-- cleanup the mess
	wipe(buffStates)
	buffStates = nil
	wipe(groups)
	groups = nil
end

local function checkForBoss()
	-- TODO implement
	return false
end

function mod:PLAYER_REGEN_ENABLED()
	if checkForBoss() then
		self:ScheduleTimer("READY_CHECK", 10)
	end
end

local dbDefaults = {
	profile = {
		state = 0,
		threshold = 10*60, -- 10 minutes
		groups = { auto = true },
		channel = "RAID",
	}
}
for i = 1,8 do
	dbDefaults.profile.groups[i] = true
end

function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("BigBrother_ReadyDB", dbDefaults)
	db = self.db.profile
	self:SetIcon()
end

function mod:OnEnable()
	debug("enabling")
	if db.state == 0 then
		return
	end
	self:SetupEvents()
end

function mod:SetupEvents()
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateMembers")
end

function mod:Resume()
	self:SetupEvents()
end

function mod:Suspend()
	self:UnregisterAllEvents()
	self:CancelAllTimers()
end

function mod:OnDisable()
	self:Suspend()
	self:Resync()
end

local stateIcons = {
	[0] = [[Interface\RAIDFRAME\ReadyCheck-NotReady]],
	[1] = [[Interface\RAIDFRAME\ReadyCheck-Ready]],
	[2] = [[Interface\RAIDFRAME\ReadyCheck-Ready]],
	-- [2] = TODO aggressive icon
}

local stateTexts = {
	[0] = "Off",
	[1] = "On",
	[2] = "On|cffff00ff/w|r",
}

local ldbdata
function mod:SetIcon()
	ldbdata.icon = stateIcons[db.state]
	ldbdata.text = stateTexts[db.state]
end

function mod:SetState(state)
	mod:SetIcon()
	if state == 0 then
		mod:Suspend()
	elseif state == 1 then
		mod:Resume()
	end
	mod:Resync()
end

local function incrementState()
	db.state = (db.state + 1) % (#stateIcons + 1)
	debug("state is now ", db.state)
	mod:SetState(db.state)
end
local stateDescriptions = {
	[0] = "is |cffff0000off|r",
	[1] = "is |cff00ff00on|r",
	[2] = "is |cff00ff00on|r and will send |cffff00ffwhipsers|r"
}
local speaker
ldbdata = {
	type = "launcher",
	OnClick = function(_, msg)
			if msg == "LeftButton" then
				if IsShiftKeyDown() then
					mod:READY_CHECK()
				else
					incrementState()
				end
			elseif msg == "RightButton" then
				mod:ShowConfig()
			end
		end,
	OnTooltipShow = function(tt)
			tt:AddLine("BigBrother - Ready")
			tt:AddLine(stateDescriptions[db.state])
			tt:AddLine("|cffffff00Click|r switch on/off")
			tt:AddLine("|cffffff00Shift-Click|r force buff-check")
			tt:AddLine("|cffffff00Right-Click|r open options")
			if speaker then
				tt:AddLine("|cff00ff00 Speaker:|r " .. speaker)
			end
		end,
	label = "BigBrother_Ready",
}

LibStub("LibDataBroker-1.1"):NewDataObject("BigBrother_Ready", ldbdata)

-- sync-stuff
local candidates = {}
local pedantic = false
local ADDON_MSG_PREFIX = "BBReady"
local playerName = UnitName("player")

function mod:Resync()
	debug("resyncing")
	SendAddonMessage(ADDON_MSG_PREFIX, "resync", "RAID")
end

local function monitorSpeaker(self)
	if not UnitIsConnected(speaker) then
		debug("speaker went offline")
		mod:CheckGroup()
		mod:CancelTimer(timers.monitorSpeaker, true)
		timers.monitorSpeaker = nil
	end
end


function mod:ChooseCandidate()
	debug("timer expired")
	if #candidates > 0 then
		table.sort(candidates)
		debug(candidates)
		speaker = candidates[1]
		if playerName == speaker then
			debug("yesss, it's me")
			self:RegisterEvent("READY_CHECK")
		else
			debug(":( it's not me")
			self:UnregisterEvent("READY_CHECK")
			timers.monitorSpeaker = self:ScheduleRepeatingTimer(monitorSpeaker, 1)
		end
	end
	debug(speaker, " was chosen")
	wipe(candidates)
	pedantic = false
	timers.ChooseCandidate = nil
end

local function sendPing()
	if db.channel ~= "RAID" then
		return
	end
	debug("ping the raid")
	local message = db.state == 2 and "pingping" or "ping"
	SendAddonMessage(ADDON_MSG_PREFIX, message, "RAID")
	timers.pingTimer = nil
end

local inRaid = false
function mod:UpdateMembers()
	local newState = not not UnitInRaid("player")
	debug(newState, "<->", inRaid)
	if newState ~= inRaid then
		debug("changed RAID-state")
		inRaid = newState
		self:Resync()
		return
	end
	if not UnitInRaid(speaker) then
		debug("the SPEAKER left the raid")
		self:CheckGroup()
	end
end

function mod:CheckGroup(arg, ...)
	pedantic = false
	if timers.ChooseCandidate then
		debug("reschedule")
		wipe(candidates)
		self:CancelTimer(timers.ChooseCandidate, true)
	end
	debug("schedule timer")
	timers.ChooseCandidate = self:ScheduleTimer("ChooseCandidate", 10)
	self:CancelTimer(timers.sendping, true)
	timers.sendping = self:ScheduleTimer(sendPing, 2)
end
function mod:CHAT_MSG_ADDON(event, prefix, message, _, sender)
	--debug(prefix, " ", message, " ", sender)
	if prefix == ADDON_MSG_PREFIX then
		debug("message: ", message, " (", sender, ")")
		if message == "pingping" then
			if not pedantic then
				wipe(candidates)
				pedantic = true
			end
			table.insert(candidates, sender)
		elseif message == "ping" and not pedantic then
			table.insert(candidates, sender)
		elseif message == "resync" then
			debug("recieved RESYNC from ", sender)
			self:CheckGroup()
		end
	end
end

local channelChoices = {
	RAID = "Raid",
	YELL = "Yell",
	SAY = "Say",
	SELF = "Self",
	-- CUSTOM = "Custom",
}
local function getChannel()
	return db.channel
end
local function setChannel(info, value)
	db.channel = value
	mod:Resync()
end

local groupChoices = {}
for i=1,8 do
	groupChoices[i] = i
end

local function getIncludeGroup(info, key)
	return db.groups[key]
end

local function setIncludeGroup(info, key, value)
	db.groups[key] = value
end

local options = {
	name = "BigBrother - Ready",
	desc = "BigBrother - Ready",
	type = 'group',
	icon = [[Interface\GROUPFRAME\UI-Group-LeaderIcon]],
	args = {
		threshold = {
			name = "Expiration Threshold",
			desc = "Expiration Threshold",
			type = 'range',
			min = 0,
			max = 60,
			step = 1,
			set = function(info, value) db.threshold = value*60 end,
			get = function(info) return db.threshold/60 end,
		},
		channel = {
			name = "Channel",
			desc = "Which channel to broadcast to",
			type = 'select',
			values = channelChoices,
			set = setChannel,
			get = getChannel,
		},
		includeGroups = {
			name = "Groups", -- TODO L[
			desc = "Only include these groups", -- TODO L[
			type = 'group',
			inline = true,
			args = {
				auto = {
					type = 'toggle',
					name = "Auto",
					desc = "Determine Groups automatically",
					get = function(info) return db.groups.auto end,
					set = function(info, value) db.groups.auto = value end,
				},
				includeGroups = {
					type = 'multiselect',
					name = "Groups",
					desc = "Specifiy groups to include in the check",
					values = groupChoices,
					get = getIncludeGroup,
					set = setIncludeGroup,
					disabled = function() return db.groups.auto end,
				},
			},
		},
	},
}

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
AceConfig:RegisterOptionsTable("BBReady", options)
AceConfigDialog:SetDefaultSize("BBReady", 500, 350)
function mod:ShowConfig()
	AceConfigDialog:Open("BBReady")
end